/*
 * Copyright (c) 2015, 2017, Oracle and/or its affiliates. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License, version 2.0,
 * as published by the Free Software Foundation.
 *
 * This program is also distributed with certain software (including
 * but not limited to OpenSSL) that is licensed under separate terms, as
 * designated in a particular file or component or in included license
 * documentation.  The authors of MySQL hereby grant you an additional
 * permission to link the program and your derivative works with the
 * separately licensed software that they have included with MySQL.
 * This program is distributed in the hope that it will be useful,  but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 * the GNU General Public License, version 2.0, for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */
#include "modules/devapi/mod_mysqlx_table_delete.h"
#include <memory>
#include <string>
#include <vector>
#include "db/mysqlx/mysqlx_parser.h"
#include "modules/devapi/mod_mysqlx_resultset.h"
#include "modules/devapi/mod_mysqlx_table.h"
#include "scripting/common.h"
#include "mysqlshdk/libs/utils/profiling.h"

using namespace std::placeholders;
using namespace mysqlsh::mysqlx;
using namespace shcore;

TableDelete::TableDelete(std::shared_ptr<Table> owner)
    : Table_crud_definition(std::static_pointer_cast<DatabaseObject>(owner)) {
  message_.mutable_collection()->set_schema(owner->schema()->name());
  message_.mutable_collection()->set_name(owner->name());
  message_.set_data_model(Mysqlx::Crud::TABLE);
  // Exposes the methods available for chaining
  add_method("delete", std::bind(&TableDelete::remove, this, _1), "data");
  add_method("where", std::bind(&TableDelete::where, this, _1), "data");
  add_method("orderBy", std::bind(&TableDelete::order_by, this, _1), "data");
  add_method("limit", std::bind(&TableDelete::limit, this, _1), "data");
  add_method("bind", std::bind(&TableDelete::bind, this, _1), "data");

  // Registers the dynamic function behavior
  register_dynamic_function(F::delete_, F::_empty);
  register_dynamic_function(F::where, F::delete_);
  register_dynamic_function(F::orderBy, F::delete_ | F::where);
  register_dynamic_function(F::limit, F::delete_ | F::where | F::orderBy);
  register_dynamic_function(F::bind, F::where | F::orderBy | F::limit | F::bind);
  register_dynamic_function(F::execute, F::delete_ | F::where | F::orderBy | F::limit | F::bind);
  register_dynamic_function(F::__shell_hook__, F::delete_ | F::where | F::orderBy | F::limit | F::bind);

  // Initial function update
  update_functions(F::_empty);
}

/**
* Initializes this record deletion handler.
* \return This TableDelete object.
*
* This function is called automatically when Table.delete() is called.
*
* The actual deletion of the records will occur only when the execute method is
* called.
*
* #### Method Chaining
*
* After this function invocation, the following functions can be invoked:
*
* - where(String searchCriteria)
* - orderBy(List sortExprStr)
* - limit(Integer numberOfRows)
* - bind(String name, Value value)
* - execute().
*
* \sa Usage examples at execute().
*/
#if DOXYGEN_JS
TableDelete TableDelete::delete () {}
#elif DOXYGEN_PY
TableDelete TableDelete::delete () {}
#endif
shcore::Value TableDelete::remove(const shcore::Argument_list &args) {
  // Each method validates the received parameters
  args.ensure_count(0, get_function_name("delete").c_str());

  std::shared_ptr<Table> table(std::static_pointer_cast<Table>(_owner));

  if (table) {
    try {
      // Updates the exposed functions
      update_functions(F::delete_);
    }
    CATCH_AND_TRANSLATE_CRUD_EXCEPTION(get_function_name("delete"));
  }

  return Value(std::static_pointer_cast<Object_bridge>(shared_from_this()));
}

//! Sets the search condition to filter the records to be deleted from the owner
//! Table.
#if DOXYGEN_CPP
//! \param args should contain an optional string expression to filter the
//! records to be deleted.
#else
//! \param searchCondition: An optional expression to filter the records to be
//! deleted.
#endif
/**
* if not specified all the records will be deleted from the table unless a limit
* is set.
* \return This TableDelete object.
*
* The searchCondition supports \a [Parameter Binding](param_binding.html).
*
* #### Method Chaining
*
* This function can be invoked only once after:
*
* - delete()
*
* After this function invocation, the following functions can be invoked:
*
* - orderBy(List sortExprStr)
* - limit(Integer numberOfRows)
* - bind(String name, Value value)
* - execute()
*
* \sa Usage examples at execute().
*/
#if DOXYGEN_JS
TableDelete TableDelete::where(String searchCondition) {}
#elif DOXYGEN_PY
TableDelete TableDelete::where(str searchCondition) {}
#endif
shcore::Value TableDelete::where(const shcore::Argument_list &args) {
  // Each method validates the received parameters
  args.ensure_count(1, get_function_name("where").c_str());

  std::shared_ptr<Table> table(std::static_pointer_cast<Table>(_owner));

  if (table) {
    try {
      message_.set_allocated_criteria(::mysqlx::parser::parse_table_filter(
          args.string_at(0), &_placeholders));

      // Updates the exposed functions
      update_functions(F::where);
    }
    CATCH_AND_TRANSLATE_CRUD_EXCEPTION(get_function_name("where"));
  }

  return Value(std::static_pointer_cast<Object_bridge>(shared_from_this()));
}

//! Sets the order in which the deletion should be done.
#if DOXYGEN_CPP
//! \param args should contain a list of expression strings defining a sort
//! criteria, the deletion will be done following the order defined by this
//! criteria.
#else
//! \param sortExprStr: A list of expression strings defining a sort criteria,
//! the deletion will be done following the order defined by this criteria.
#endif
/**
* \return This TableDelete object.
*
* The elements of sortExprStr list are strings defining the column name on which
* the sorting will be based in the form of "columnIdentifier [ ASC | DESC ]".
* If no order criteria is specified, ascending will be used by default.
*
* This method is usually used in combination with limit to fix the amount of
* records to be deleted.
*
* #### Method Chaining
*
* This function can be invoked only once after:
*
* - delete()
* - where(String searchCondition)
*
* After this function invocation, the following functions can be invoked:
*
* - limit(Integer numberOfRows)
* - bind(String name, Value value)
* - execute()
*
*/
#if DOXYGEN_JS
TableDelete TableDelete::orderBy(List sortExprStr) {}
#elif DOXYGEN_PY
TableDelete TableDelete::order_by(list sortExprStr) {}
#endif
shcore::Value TableDelete::order_by(const shcore::Argument_list &args) {
  args.ensure_at_least(1, get_function_name("orderBy").c_str());

  try {
    std::vector<std::string> fields;

    parse_string_list(args, fields);

    if (fields.size() == 0)
      throw shcore::Exception::argument_error(
          "Order criteria can not be empty");

    for (auto &field : fields)
      ::mysqlx::parser::parse_table_sort_column(*message_.mutable_order(),
                                                field);

    update_functions(F::orderBy);
  }
  CATCH_AND_TRANSLATE_CRUD_EXCEPTION(get_function_name("orderBy"));

  return Value(std::static_pointer_cast<Object_bridge>(shared_from_this()));
}

//! Sets a limit for the records to be deleted.
#if DOXYGEN_CPP
//! \param args should contain the number of records to be deleted.
#else
//! \param numberOfRows the number of records to be deleted.
#endif
/**
* \return This TableDelete object.
*
* This method is usually used in combination with sort to fix the amount of
* records to be deleted.
*
* #### Method Chaining
*
* This function can be invoked only once after:
*
* - delete()
* - where(String searchCondition)
* - orderBy(List sortExprStr)
*
* After this function invocation, the following functions can be invoked:
*
* - bind(String name, Value value)
* - execute()
*
* \sa Usage examples at execute().
*/
#if DOXYGEN_JS
TableDelete TableDelete::limit(Integer numberOfRows) {}
#elif DOXYGEN_PY
TableDelete TableDelete::limit(int numberOfRows) {}
#endif
shcore::Value TableDelete::limit(const shcore::Argument_list &args) {
  args.ensure_count(1, get_function_name("limit").c_str());

  try {
    message_.mutable_limit()->set_row_count(args.uint_at(0));

    update_functions(F::limit);
  }
  CATCH_AND_TRANSLATE_CRUD_EXCEPTION(get_function_name("limit"));

  return Value(std::static_pointer_cast<Object_bridge>(shared_from_this()));
}

//! Binds a value to a specific placeholder used on this TableDelete object.
#if DOXYGEN_CPP
//! \param args should contain the next elements:
//! \li The name of the placeholder to which the value will be bound.
//! \li The value to be bound on the placeholder.
#else
//! \param name: The name of the placeholder to which the value will be bound.
//! \param value: The value to be bound on the placeholder.
#endif
/**
* \return This TableDelete object.
*
* #### Method Chaining
*
* This function can be invoked multiple times right before calling execute:
*
* After this function invocation, the following functions can be invoked:
*
* - bind(String name, Value value)
* - execute()
*
* An error will be raised if the placeholder indicated by name does not exist.
*
* This function must be called once for each used placeohlder or an error will
* be
* raised when the execute method is called.
*
* \sa Usage examples at execute().
*/
#if DOXYGEN_JS
TableDelete TableDelete::bind(String name, Value value) {}
#elif DOXYGEN_PY
TableDelete TableDelete::bind(str name, Value value) {}
#endif
shcore::Value TableDelete::bind(const shcore::Argument_list &args) {
  args.ensure_count(2, get_function_name("bind").c_str());

  try {
    bind_value(args.string_at(0), args[1]);

    update_functions(F::bind);
  }
  CATCH_AND_TRANSLATE_CRUD_EXCEPTION(get_function_name("bind"));

  return Value(std::static_pointer_cast<Object_bridge>(shared_from_this()));
}

/**
* Executes the record deletion with the configured filter and limit.
* \return Result A result object that can be used to retrieve the results of the
* deletion operation.
*
* #### Method Chaining
*
* This function can be invoked after any other function on this class.
*/
#if DOXYGEN_JS
/**
*
* #### Examples
* \dontinclude "js_devapi/scripts/mysqlx_table_delete.js"
* \skip //@ TableDelete: delete under condition
* \until //@ TableDelete: with limit 3
* \until print('Records Left:', records.length, '\n');
*/
Result TableDelete::execute() {}
#elif DOXYGEN_PY
/**
*
* #### Examples
* \dontinclude "py_devapi/scripts/mysqlx_table_delete.py"
* \skip #@ TableDelete: delete under condition
* \until #@ TableDelete: with limit 3
* \until print 'Records Left:', len(records), '\n'
*/
Result TableDelete::execute() {}
#endif
shcore::Value TableDelete::execute(const shcore::Argument_list &args) {
  std::unique_ptr<mysqlsh::mysqlx::Result> result;
  args.ensure_count(0, get_function_name("execute").c_str());
  try {
    mysqlshdk::utils::Profile_timer timer;
    insert_bound_values(message_.mutable_args());
    timer.stage_begin("TableDelete::execute");
    result.reset(new mysqlsh::mysqlx::Result(safe_exec(
        [this]() { return session()->session()->execute_crud(message_); })));
    timer.stage_end();
    result->set_execution_time(timer.total_seconds_ellapsed());
  }
  CATCH_AND_TRANSLATE_CRUD_EXCEPTION(get_function_name("execute"));

  return result ? shcore::Value::wrap(result.release()) : shcore::Value::Null();
}
